// SilentTimer Action Figure communication library
// libstaf.c
// 
// 27.06.2020
// Copyright (C) 2020 Balthasar Szczepański
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <fcntl.h>
#include <unistd.h>
#include <termios.h>

// #include <stdio.h>//

#include "libstaf.h"

#define STAF_BAUD B9600
#define STAF_BUFFER_SIZE 32

#define CMD_GET_TIME        'G'
#define CMD_GET_MODE        'H'
#define CMD_GET_DISPLAY     'I'
#define CMD_GET_VALUE       'J'
#define CMD_GET_CALCULATION 'K'
#define CMD_GET_RUN_MODE    'L'
#define CMD_GET_VERSION     'M'
#define CMD_GET_PINOUT      'N'
#define CMD_GET_EVENT       'O'
#define CMD_GET_ALARM       'P'
#define CMD_GET_DIGIT       'Q'
#define CMD_READ_MEMORY     'R'

#define CMD_SET_DISPLAY       'g'
#define CMD_FREE_DISPLAY      'h'
#define CMD_TRIGGER_EVENT     'i'
#define CMD_GOTO_MODE         'j'
#define CMD_SELECT_DIGIT      'k'
#define CMD_SET_DIGIT_VALUE   'l'
#define CMD_CONFIRM           'm'
#define CMD_SET_TIME          'n'
#define CMD_SET_ALARM         'o'
#define CMD_SET_DATE          'p'
#define CMD_SET_YEAR          'q'
#define CMD_SET_COUNTDOWN     'r'
#define CMD_SET_CALIBRATION   's'
#define CMD_SET_DEBUG_ADDRESS 't'
#define CMD_SET_DEBUG_VALUE   'u'
#define CMD_CALCULATE         'v'
#define CMD_SET_VALUE         'w'
#define CMD_SET_RUN_MODE      'x'
#define CMD_SET_BUZZER_MODE   'y'
#define CMD_WRITE_MEMORY      'z'

#define CMD_TYPE 0x20


#define CALC_NEGATIVE 0x01 
#define CALC_FRACTION 0x02 


uint8_t stafEncodeBcdU8 (uint8_t x);
uint8_t stafDecodeBcdU8 (uint8_t x);
uint16_t stafEncodeBcdU16 (uint16_t x);
uint16_t stafDecodeBcdU16 (uint16_t x);
int stafEncodeCalculationArgument (float in, uint8_t *out, uint8_t *flags);
float stafDecodeCalculationResult (uint8_t *in, uint8_t flags);

ssize_t stafWrite (int fd, uint8_t *buffer, size_t length);
ssize_t stafRead (int fd, uint8_t *buffer, size_t length);

int stafSendFrame (int fd, uint8_t command, uint8_t *data, size_t length);
int stafGetFrame (int fd, uint8_t *command, uint8_t *data, size_t length);

int stafExecuteCommand (int fd, uint8_t *event, uint8_t command, uint8_t *dataOut, uint8_t *dataIn, size_t lengthOut, size_t lengthIn);


uint8_t stafEncodeBcdU8 (uint8_t x)
{
	uint8_t p;
	uint8_t q;
	uint8_t y;
	
	p = x/10;
	q = x%10;
	
	y = (p<<4) | q;
	
	return y;
}

uint8_t stafDecodeBcdU8 (uint8_t x)
{
	uint8_t p;
	uint8_t q;
	uint8_t y;
	
	p = x>>4;
	q = x&0x0f;
	
	y = p*10+q;
	
	return y;
}

uint16_t stafEncodeBcdU16 (uint16_t x)
{
	uint8_t p;
	uint8_t q;
	uint16_t y;
	
	p = (uint8_t)(x/100);
	q = (uint8_t)(x%100);
	
	y = (stafEncodeBcdU8(p)<<8)|stafEncodeBcdU8(q);
	
	return y;
}

uint16_t stafDecodeBcdU16 (uint16_t x)
{
	uint8_t p;
	uint8_t q;
	uint16_t y;
	
	p = (uint8_t)(x>>8);
	q = (uint8_t)(x&0xff);
	
	y = (stafDecodeBcdU8(p)*100)+stafDecodeBcdU8(q);
	
	return y;
}


//serial interfacing code (especially the configuration) used rl78flash by Maksim Salau (MIT license) as an example of how to do it.

int stafConnect(const char *portname)
{
	int fd;
	struct termios opt;
	
	fd = open(portname, O_RDWR | O_NOCTTY | O_NDELAY);
	if (fd != -1)
	{
		fcntl(fd, F_SETFL, 0);
		tcgetattr(fd, &opt);
		cfsetospeed(&opt, STAF_BAUD);
		cfsetispeed(&opt, STAF_BAUD);
		opt.c_cflag &= ~(HUPCL | CSIZE | PARENB | CRTSCTS | CSTOPB);
		opt.c_cflag |= CLOCAL | CREAD | CS8;
		opt.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
		opt.c_iflag &= ~(IXON | IXOFF | IXANY);
		opt.c_oflag &= ~OPOST;
		opt.c_cc[VMIN] = 0;
		opt.c_cc[VTIME] = 1;
		tcsetattr(fd, TCSANOW, &opt);
		usleep(1000);
		tcflush(fd,TCIOFLUSH);
	}
	return fd;
}

int stafDisconnect(int fd)
{
	return close(fd);
}


ssize_t stafRead (int fd, uint8_t *buffer, size_t length)
{
	size_t done;
	ssize_t r;
	uint8_t *point;
	uint8_t inbyte;
	
	done = 0;
	point = buffer;
	while (done < length)
	{
		r = read(fd, &inbyte, 1);
		if (r<0)
			return r;
		if (r==0)
			break;
		
		if (inbyte == 'U')
			continue;
		if (((inbyte >='G') && (inbyte <='Z')) || ((inbyte >='g') && (inbyte <='z')))
		{
			point = buffer;
			done = 0;
		}
		
		*point = inbyte;
		++point;
		++done;
		
		if ((inbyte == '\r') || (inbyte == '\n'))
		{
			if (done == 1)
			{
				point = buffer;
				done = 0;
				continue;
			}
			else
				break;
		}
	}
	return (ssize_t) done;
}

ssize_t stafWrite (int fd, uint8_t *buffer, size_t length)
{
	size_t left;
	ssize_t r;
	uint8_t *point;
	
	left = length;
	point = buffer;
	
	while (left > 0)
	{
		r = write(fd, point, left);
		if (r < 0)
			return r;
		point += r;
		left -= r;
	}
	return (ssize_t)(length-left);
}


int stafGetFrame (int fd, uint8_t *command, uint8_t *data, size_t length)
{
	uint8_t buffer[STAF_BUFFER_SIZE];
	ssize_t r;
	size_t i,j;
	uint8_t temp;
	
	if((2*(length+1))>STAF_BUFFER_SIZE)
		return ERR_DATA_TOO_LONG;
	
	r = stafRead(fd, buffer, STAF_BUFFER_SIZE);
	if (r<0)
		return (int) r;
	
	if((r==0)&&(length!=0))
		return ERR_NO_DATA;
	if (r<(2*(length+1)))
		return ERR_INCOMPLETE_DATA;
	if ((buffer[(2*length)+1]!='\r')&&(buffer[(2*length)+1]!='\n'))
		return ERR_INVALID_DATA;
	if (!(((buffer[0]>='G')&&(buffer[0]<='Z'))||((buffer[0]>='g')&&(buffer[0]<='z'))))
		return ERR_INVALID_DATA;
	
	*command = buffer[0];
	
	for (i=0,j=1; i<length; ++j)
	{
		if ((buffer[j]>='0')&&(buffer[j]<='9'))
			temp = buffer[j] - '0';
		else if ((buffer[j]>='A')&&(buffer[j]<='F'))
			temp = buffer[j] - ('A'-10);
		else if ((buffer[j]>='a')&&(buffer[j]<='f'))
			temp = buffer[j] - ('a'-10);
		else
			return ERR_INVALID_DATA;
		
		if (j & 1)
			data[i] = temp<<4;
		else
		{
			data[i] |= temp;
			++i;
		}
	}
	return 0;
}

int stafSendFrame (int fd, uint8_t command, uint8_t *data, size_t length)
{
	uint8_t buffer[STAF_BUFFER_SIZE];
	ssize_t r;
	size_t i,j;
	uint8_t temp;
	
	if((2*(length+1))>STAF_BUFFER_SIZE)
		return ERR_DATA_TOO_LONG;
	
	if (!(((command>='G')&&(command<='Z'))||((command>='g')&&(command<='z'))))
		return ERR_INVALID_DATA;
	
	buffer[0]=command;
	
	for (i=0,j=1; i<length; ++j)
	{
		if (j & 1)
			temp = data[i]>>4;
		else
		{
			temp = data[i]&0x0f;
			++i;
		}
		if (temp < 10)
			buffer[j] = temp + '0';
		else
			buffer[j] = temp + 'a' - 10;
	}
	buffer[j] = '\n';
	
	r = stafWrite(fd, buffer, 2*(length+1));
	if (r<0)
		return (int) r;
	
	if (r<(2*(length+1)))
		return ERR_INCOMPLETE_DATA;
	
	return 0;
}


int stafExecuteCommand (int fd, uint8_t *event, uint8_t command, uint8_t *dataOut, uint8_t *dataIn, size_t lengthOut, size_t lengthIn)
{
	int status=0;
	uint8_t temp;
	
	if (event!=NULL)
	{
		status = stafCheckEvent(fd, event);
		if(status != 0)
			return status;
	}
	
	tcflush(fd,TCIOFLUSH);
	
	status = stafSendFrame(fd, command, dataOut, lengthOut);
	if(status != 0)
		return status;
	
	if (!(command & CMD_TYPE))
	{
		status = stafGetFrame(fd, &temp, dataIn, lengthIn);
		if(status != 0)
			return status;
		
		if(temp != command)
			return ERR_INVALID_DATA;
	}
	return 0;
}


int stafGetState (int fd, uint8_t *event, struct StafState *state)
{
	int status = 0;
	
	if (state==NULL)
		return ERR_MISSING_POINTER;
	
	status = stafGetMode(fd, event, &(state->mode));
	if(status != 0)
		return status;
	
	status = stafIsRunning(fd, event, &(state->running));
	if(status != 0)
		return status;
	
	status = stafGetDigit(fd, event, &(state->digit));
	if(status != 0)
		return status;
	
	status = stafGetValue(fd, event, state->value);
	if(status != 0)
		return status;
	
	return 0;
}

int stafSetState (int fd, uint8_t *event, struct StafState *state)
{
	int status = 0;
	
	if (state==NULL)
		return ERR_MISSING_POINTER;
	
	status = stafSetMode(fd, event, state->mode);
	if(status != 0)
		return status;
	
	status = stafSetRunning(fd, event, state->running);
	if(status != 0)
		return status;
	
	status = stafSetDigit(fd, event, state->digit);
	if(status != 0)
		return status;
	
	status = stafSetValue(fd, event, state->value);
	if(status != 0)
		return status;
	
	return 0;
}


int stafGetMode (int fd, uint8_t *event, uint8_t *mode)
{
	int status = 0;
	uint8_t indata;
	
	status = stafExecuteCommand(fd, event, CMD_GET_MODE, NULL, &indata, 0, 1);
	if (status != 0)
		return status;
	
	if (mode != NULL)
		*mode = indata;
	
	return 0;
}

int stafSetMode (int fd, uint8_t *event, uint8_t mode)
{
	int status = 0;
	
	if(!((mode>=mode_time)&&(mode<=mode_calculator))||(mode>=mode_set_time)&&(mode<=mode_set_calc_operator))
		return ERR_INVALID_PARAMETER;
	
	status = stafExecuteCommand(fd, event, CMD_GOTO_MODE, &mode, NULL, 1, 0);
	if (status != 0)
		return status;
	
	return 0;
}


int stafCheckEvent (int fd, uint8_t *event)
{
	int status = 0;
	uint8_t command;
	uint8_t data;
	
	while(1)
	{
		status = stafGetFrame(fd,&command,&data,1);
		
		if (status == 0)
		{
			if (command == CMD_GET_EVENT)
				*event |= data;
			else
				return ERR_INVALID_DATA;
		}
		else if (status == ERR_NO_DATA)
			return 0;
		else
			return status;
	}
}

int stafSubscribeEvent (int fd, uint8_t *event, uint8_t subscribeEvent)
{
	int status = 0;
	uint8_t indata;
	uint8_t outdata;
	
	outdata = subscribeEvent & event_mask;
	
	status = stafExecuteCommand(fd, event, CMD_GET_EVENT, &outdata, &indata, 1, 1);
	if (status != 0)
		return status;
	
	if (event != NULL)
	{
		*event |= indata;
	}
	
	return 0;
}

int stafTriggerEvent (int fd, uint8_t *event, uint8_t triggerEvent)
{
	int status = 0;
	uint8_t outdata;
	
	outdata = triggerEvent & event_mask;
	
	status = stafExecuteCommand(fd, event, CMD_TRIGGER_EVENT, &outdata, NULL, 1, 0);
	
	if (status != 0)
		return status;
	
	return 0;
}


int stafGetPinout (int fd, uint8_t *event, uint8_t *pinout)
{
	int status = 0;
	uint8_t indata[4];
	int i,j;
	
	status = stafExecuteCommand(fd, event, CMD_GET_PINOUT, NULL, indata, 0, 4);
	if (status != 0)
		return status;
	
	if (pinout != NULL)
	{
		for (i=0,j=0; i<4; ++i)
		{
			pinout[j] = ~(1<<(indata[i]>>4));
			++j;
			pinout[j] = ~(1<<(indata[i]&0x0f));
			++j;
		}
	}
	
	return 0;
}

int stafGetDisplay (int fd, uint8_t *event, uint8_t *pinout, uint32_t *display)
{
	int status = 0;
	uint8_t indata[4];
	int i,j;
	
	status = stafExecuteCommand(fd, event, CMD_GET_DISPLAY, NULL, indata, 0, 4);
	if (status != 0)
		return status;
	
	if (display != NULL)
	{
		*display = 0;
		for(i=0; i<4; ++i)
		{
			for (j=0; j<8; ++j)
			{
				if ((indata[i]|pinout[j])!=0xff)
					*display |= 1<<(i*8+j);
			}
		}
	}
	
	return 0;
}

int stafSetDisplay (int fd, uint8_t *event, uint8_t *pinout, uint32_t display)
{
	int status = 0;
	uint8_t outdata[4];
	int i,j;
	
	for(i=0; i<4; ++i)
	{
		outdata[i]=0xff;
		for (j=0; j<8; ++j)
		{
			if (display & (1<<(i*8+j)))
				outdata[i] &= pinout[j];
		}
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_DISPLAY, outdata, NULL, 4, 0);
	if (status != 0)
		return status;
	
	return 0;
}

int stafFreeDisplay (int fd, uint8_t *event)
{
	int status = 0;
	
	status = stafExecuteCommand(fd, event, CMD_FREE_DISPLAY, NULL, NULL, 0, 0);
	if (status != 0)
		return status;
	
	return 0;
}


int stafGetTime (int fd, uint8_t *event, uint8_t *cs, uint8_t *second, uint8_t *minute, uint8_t *hour, uint8_t *day, uint8_t *month, uint16_t *year)
{
	int status = 0;
	uint8_t indata[8];
	
	status = stafExecuteCommand(fd, event, CMD_GET_TIME, NULL, indata, 0, 8);
	if (status != 0)
		return status;
	
	if (cs != NULL)
		*cs = stafDecodeBcdU8(indata[0]);
	if (second != NULL)
		*second = stafDecodeBcdU8(indata[1]);
	if (minute != NULL)
		*minute = stafDecodeBcdU8(indata[2]);
	if (hour != NULL)
		*hour = stafDecodeBcdU8(indata[3]);
	if (day != NULL)
		*day = stafDecodeBcdU8(indata[4]);
	if (month != NULL)
		*month = stafDecodeBcdU8(indata[5]);
	if (year != NULL)
		*year = stafDecodeBcdU16((((uint16_t)indata[7])<<8)+indata[6]);
	
	return 0;
}

int stafSetTime (int fd, uint8_t *event, uint8_t minute, uint8_t hour, uint8_t restoreState)
{
	int status = 0;
	uint8_t outdata[2];
	struct StafState state;
	
	if (!((hour<=23)&&(minute<=59)))
		return ERR_INVALID_PARAMETER;
	
	outdata[0] = stafEncodeBcdU8(minute);
	outdata[1] = stafEncodeBcdU8(hour);
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_TIME, outdata, NULL, 2, 0);
	if (status != 0)
		return status;
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}

int stafSetDate (int fd, uint8_t *event, uint8_t day, uint8_t month, uint8_t restoreState)
{
	int status = 0;
	uint8_t outdata[2];
	struct StafState state;
	
	if (!((day>=1)&&(day<=31)&&(month>=1)&&(month<=12)))
		return ERR_INVALID_PARAMETER;
	
	outdata[0] = stafEncodeBcdU8(day);
	outdata[1] = stafEncodeBcdU8(month);
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_DATE, outdata, NULL, 2, 0);
	if (status != 0)
		return status;
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}

int stafSetYear (int fd, uint8_t *event, uint16_t year, uint8_t restoreState)
{
	int status = 0;
	uint8_t outdata[2];
	struct StafState state;
	uint16_t yearbcd;
	
	if (year>9999)
		return ERR_INVALID_PARAMETER;
	
	yearbcd = stafEncodeBcdU16(year);
	outdata[0] = yearbcd&0xff;
	outdata[1] = yearbcd>>8;
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_YEAR, outdata, NULL, 2, 0);
	if (status != 0)
		return status;
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}


int stafGetAlarm (int fd, uint8_t *event, uint8_t *enabled, uint8_t *minute, uint8_t *hour)
{
	int status = 0;
	uint8_t indata[2];
	
	status = stafExecuteCommand(fd, event, CMD_GET_ALARM, NULL, indata, 0, 2);
	if (status != 0)
		return status;
	
	if (enabled != NULL)
		*enabled = ((indata[0]==0xff)&&(indata[1]==0xff))?0:1;
	if (minute != NULL)
		*minute = stafDecodeBcdU8(indata[0]);
	if (hour != NULL)
		*hour = stafDecodeBcdU8(indata[1]);
	
	return 0;
}

int stafSetAlarm (int fd, uint8_t *event, uint8_t minute, uint8_t hour, uint8_t restoreState)
{
	int status = 0;
	uint8_t outdata[2];
	struct StafState state;
	
	if (!((hour<=23)&&(minute<=59)))
		return ERR_INVALID_PARAMETER;
	
	outdata[0] = stafEncodeBcdU8(minute);
	outdata[1] = stafEncodeBcdU8(hour);
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_ALARM, outdata, NULL, 2, 0);
	if (status != 0)
		return status;
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}

int stafClearAlarm (int fd, uint8_t *event, uint8_t restoreState)
{
	int status = 0;
	uint8_t outdata[2] = {0xff, 0xff};
	struct StafState state;
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_ALARM, outdata, NULL, 2, 0);
	if (status != 0)
		return status;
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}


int stafSetCountdown (int fd, uint8_t *event, uint8_t second, uint8_t minute, uint8_t restoreState)
{
	int status = 0;
	uint8_t outdata[2];
	struct StafState state;
	
	if (!((minute<=99)&&(second<=59)))
		return ERR_INVALID_PARAMETER;
	
	outdata[0] = stafEncodeBcdU8(second);
	outdata[1] = stafEncodeBcdU8(minute);
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_COUNTDOWN, outdata, NULL, 2, 0);
	if (status != 0)
		return status;
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}

int stafIsRunning (int fd, uint8_t *event, uint8_t *running)
{
	int status = 0;
	uint8_t indata;
	
	status = stafExecuteCommand(fd, event, CMD_GET_RUN_MODE, NULL, &indata, 0, 1);
	if (status != 0)
		return status;
	
	if (running != NULL)
		*running = indata;
	
	return 0;
}

int stafSetRunning (int fd, uint8_t *event, uint8_t running)
{
	int status = 0;
	uint8_t outdata;
	
	outdata = running ? 1 : 0;
	
	status = stafExecuteCommand(fd, event, CMD_SET_RUN_MODE, &outdata, NULL, 1, 0);
	if (status != 0)
		return status;
	
	return 0;
}


int stafGetVersion (int fd, uint8_t *event, uint8_t *minor, uint8_t *major)
{
	int status = 0;
	uint8_t indata[2];
	
	status = stafExecuteCommand(fd, event, CMD_GET_VERSION, NULL, indata, 0, 2);
	if (status != 0)
		return status;
	
	if (minor != NULL)
		*minor = stafDecodeBcdU8(indata[0]);
	if (major != NULL)
		*major = stafDecodeBcdU8(indata[1]);
	
	return 0;
}


int stafGetDigit (int fd, uint8_t *event, uint8_t *digit)
{
	int status = 0;
	uint8_t indata;
	
	status = stafExecuteCommand(fd, event, CMD_GET_DIGIT, NULL, &indata, 0, 1);
	if (status != 0)
		return status;
	
	if (digit != NULL)
		*digit = indata;
	
	return 0;
}

int stafSetDigit (int fd, uint8_t *event, uint8_t digit)
{
	int status = 0;
	
	if (digit > 4)
		return ERR_INVALID_PARAMETER;
	
	status = stafExecuteCommand(fd, event, CMD_SELECT_DIGIT, &digit, NULL, 1, 0);
	if (status != 0)
		return status;
	
	return 0;
}

int stafGetDigitValue (int fd, uint8_t *event, uint8_t *value)
{
	int status = 0;
	uint8_t indata[3];
	uint8_t digit;
	
	status = stafGetDigit(fd, event, &digit);
	if (status != 0)
		return status;
	
	status = stafGetValue(fd, event, indata);
	if (status != 0)
		return status;
	
	*value = indata[digit>>1];
	
	if(digit & 0x01)
		*value >>= 4;
	else
		*value &= 0x0f;
	
	return 0;
}

int stafSetDigitValue (int fd, uint8_t *event, uint8_t value)
{
	int status = 0;
	
	if (value > 0x0f)
		return ERR_INVALID_PARAMETER;
	
	status = stafExecuteCommand(fd, event, CMD_SET_DIGIT_VALUE, &value, NULL, 1, 0);
	if (status != 0)
		return status;
	
	return 0;
}

int stafGetValue (int fd, uint8_t *event, uint8_t *value)
{
	int status = 0;
	uint8_t indata[3];
	
	status = stafExecuteCommand(fd, event, CMD_GET_VALUE, NULL, indata, 0, 3);
	if (status != 0)
		return status;
	
	if (value != NULL)
	{
		value[0]=indata[0];
		value[1]=indata[1];
		value[2]=indata[2];
	}
	
	return 0;
}

int stafSetValue (int fd, uint8_t *event, uint8_t *value)
{
	int status = 0;
	
	if (value == NULL)
		return ERR_MISSING_POINTER;
	
	status = stafExecuteCommand(fd, event, CMD_SET_VALUE, value,  NULL, 3, 0);
	if (status != 0)
		return status;
	
	return 0;
}

int stafConfirm (int fd, uint8_t *event)
{
	int status = 0;
	
	status = stafExecuteCommand(fd, event, CMD_CONFIRM, NULL, NULL, 0, 0);
	if (status != 0)
		return status;
	
	return 0;
}


int stafSetDebugAddress (int fd, uint8_t *event, uint8_t address)
{
	int status = 0;
	
	status = stafExecuteCommand(fd, event, CMD_SET_DEBUG_ADDRESS, &address, NULL, 1, 0);
	if (status != 0)
		return status;
	
	return 0;
}

int stafGetDebugValue (int fd, uint8_t *event, uint8_t address, uint8_t *value, uint8_t restoreState)
{
	int status = 0;
	uint8_t indata[3];
	struct StafState state;
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafSetDebugAddress(fd, event, address);
	if (status != 0)
		return status;
	
	status = stafTriggerEvent(fd, event, event_tick);
	if (status != 0)
		return status;
	
	status = stafGetValue(fd, event, indata);
	if (status != 0)
		return status;
	
	*value = indata[0];
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}

int stafSetDebugValue (int fd, uint8_t *event, uint8_t address, uint8_t value, uint8_t restoreState)
{
	int status = 0;
	uint8_t outdata[2];
	struct StafState state;
	
	outdata[0] = value;
	outdata[1] = address;
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_DEBUG_VALUE, outdata, NULL, 2, 0);
	if (status != 0)
		return status;
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}

int stafReadMemory (int fd, uint8_t *event, uint16_t address, uint8_t *value)
{
	int status = 0;
	uint8_t indata;
	uint8_t outdata[2];
	
	outdata[0] = address&0xff;
	outdata[1] = address>>8;
	
	status = stafExecuteCommand(fd, event, CMD_READ_MEMORY, outdata, &indata, 2, 1);
	if (status != 0)
		return status;
	
	if (value != NULL)
		*value = indata;
	
	return 0;
}

int stafWriteMemory (int fd, uint8_t *event, uint16_t address, uint8_t value)
{
	int status = 0;
	uint8_t indata;
	uint8_t outdata[2];
	
	outdata[0] = address&0xff;
	outdata[1] = address>>8;
	outdata[2] = value;
	
	status = stafExecuteCommand(fd, event, CMD_WRITE_MEMORY, outdata, NULL, 3, 0);
	if (status != 0)
		return status;
	
	return 0;
}


int stafGetCalibration (int fd, uint8_t *event, int16_t *calibration, uint8_t restoreState)
{
	int status = 0;
	uint8_t indata[3];
	struct StafState state;
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafSetMode(fd, event, mode_calibration);
	if (status != 0)
		return status;
	
	status = stafGetValue(fd, event, indata);
	if (status != 0)
		return status;
	
	if (calibration != NULL)
	{
		*calibration = (int16_t)indata[0];
		if(indata[1])
			*calibration=0-*calibration;
	}
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}

int stafSetCalibration (int fd, uint8_t *event, int16_t calibration, uint8_t restoreState)
{
	int status = 0;
	uint8_t outdata[2];
	struct StafState state;
	
	if (!((calibration>=-255)&&(calibration<=255)))
		return ERR_INVALID_PARAMETER;
	
	if(calibration<0)
	{
		outdata[0] = (uint8_t) (0-calibration);
		outdata[1] = 1;
	}
	else
	{
		outdata[0] = (uint8_t) calibration;
		outdata[1] = 0;
	}
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	status = stafExecuteCommand(fd, event, CMD_SET_CALIBRATION, outdata, NULL, 2, 0);
	if (status != 0)
		return status;
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}


int stafSetBuzzerMode (int fd, uint8_t *event, uint8_t mode)
{
	int status = 0;
	
	if (mode > buzzer_forced)
		return ERR_INVALID_PARAMETER;
	
	status = stafExecuteCommand(fd, event, CMD_SET_BUZZER_MODE, &mode, NULL, 1, 0);
	if (status != 0)
		return status;
	
	return 0;
}


int stafEncodeCalculationArgument (float in, uint8_t *out, uint8_t *flags)
{
	*flags = 0;
	float limit, fraction;
	uint16_t bcd;
	
	if (in <0)
	{
		*flags |= CALC_NEGATIVE;
		in *= -1.0;
		limit = 999.5;
	}
	else
		limit = 9999.5;
	
	if (in>=limit)
		return ERR_INVALID_PARAMETER;
	
	fraction = in*100.0;
	if(fraction>limit)
		bcd = stafEncodeBcdU16((uint16_t)(in+0.5));
	else
	{
		bcd = stafEncodeBcdU16((uint16_t)(fraction+0.5));
		*flags |= CALC_FRACTION;
	}
	out[0]=bcd&0xff;
	out[1]=bcd>>8;
	
	return 0;
}

float stafDecodeCalculationResult (uint8_t *in, uint8_t flags)
{
	float out;
	
	out = (float)stafDecodeBcdU16((((uint16_t)in[1])<<8)|in[0]);
	
	if(flags&CALC_NEGATIVE)
		out*=-1.0;
	if(flags&CALC_FRACTION)
		out*=0.01;
	
	return out;
}

int stafCalculate (int fd, uint8_t *event, float a, float b, uint8_t operation, float *y, uint8_t *error, uint8_t restoreState)
{
	int status = 0;
	uint8_t indata[8];
	uint8_t outdata[5];
	struct StafState state;
	uint8_t flags;
	
	if (restoreState)
	{
		status = stafGetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	outdata[4] |= operation & calc_mask;
	
	status=stafEncodeCalculationArgument (b, outdata, &flags);
	if (status != 0)
		return status;
	outdata[4] |= flags<<2;
	
	status=stafEncodeCalculationArgument (a, outdata+2, &flags);
	if (status != 0)
		return status;
	outdata[4] |= flags<<4;
	
	status = stafExecuteCommand(fd, event, CMD_CALCULATE, outdata, NULL, 5, 0);
	if (status != 0)
		return status;
	
	status = stafExecuteCommand(fd, event, CMD_GET_CALCULATION, NULL, indata, 0, 8);
	if (status != 0)
		return status;
	
	*error = indata[7];
	*y = stafDecodeCalculationResult(indata,indata[6]);
	
	if (restoreState)
	{
		status = stafSetState(fd, event, &state);
		if (status != 0)
			return status;
	}
	
	return 0;
}

